/***********************************************************************
**
**	lzv.c -- an extremly fast compression/decompression-method
**
**	Written by Hermann Vogt
**
**		v 0.4 -- 96/03/30   adapted by Markus F.X.J. Oberhumer for LZO-test
**		v 0.3 -- 94/03/08	aCembler version of rLZV built in.
**		v 0.2 -- 94/03/04	Changes for usage with DouBle 0.2 built in.
**		v 0.1 -- 94/03/01	Intensivly tested, removed all known bugs.
**		v 0.0 -- 94/02/21	First Version.
**
**	Copyright (c) 1994 Hermann Vogt. Redistribution of this file is
**	permitted under the GNU Public Licence.
**
**	The method presented here is faster and compresses better
**	than lzrw1 and lzrw1-a. I named it lzv for "Lev-Zimpel-Vogt".
**	It uses ideas introduced by Ross Williams in his algorithm lzrw1
**	[R. N. Williams (1991): "An Extremly Fast ZIV-Lempel Data
**	Compression Algorithm", Proceedings IEEE Data Compression
**	Conference, Snowbird, Utah, 362-371] and by Fiala and Green in their
**	algorithm a1 [E. R. Fiala, D. H. Greene (1989): "Data Compression
**	with Finite Windows", Communications of the ACM, 4, 490-505].
**	Because lzv differs strongly from both, I hope there will be no
**	patent problems. The hashing-method has been stolen from Jean-loup
**	Gailly's (patent free) gzip.
**
**	KNOWN PROBLEMS:
**	-	My english is very bad.
**  -	Badly commented. (I hope this will be better in the next
**		version.)
**	-	I'm not sure if lzv is free from patent problems.
**
***********************************************************************/

#ifndef MFX
#define MFX
#endif

#define NDEBUG
#include <assert.h>
#include <string.h>

#include <lzv.h>

#define	MLL		32			/*	Maximum len of chain of literals	*/
#define	MML		(8+256)		/*	Maximum len of match				*/
#define	MOFF	8191		/*	Maximum offset						*/
#define	HSIZ	32768u		/*	Size of Hashtable					*/

#define uch     lzo_byte
#define uit     lzo_uint
#define ush     unsigned short __LZO_MMODEL



#if defined(MFX)
/* mfx: we needed lzo_uint for a blocksize greater than 65535 */
#define htab_t	lzo_uint
#else
typedef unsigned short htab_t;
#endif


static
int
wLZV(const uch *in, uch *out, uit len, uit *out_len, void *wrkmem)
{
	uit hval, op, ip, l_len, m_pos, m_off, m_len, maxlen;
	htab_t * const htab = wrkmem;

#if 0
	/* be deterministic */
	memset(htab,0,HSIZ*sizeof(htab_t));
#endif

	maxlen = *out_len;
	*out_len = 0;
	hval = ((in[0] << 5) ^ in[1]) & (HSIZ - 1);
	ip = op = l_len = 0;
	do {
		hval = ((hval << 5) ^ in[ip+2]) & (HSIZ - 1);
		m_pos = htab[hval];
		htab[hval] = ip;
		if (m_pos < ip &&
				in[m_pos] == in[ip] &&
				(m_off = ip - m_pos - 1) <= MOFF &&
				ip + 4 < len &&
				*(ush *)(in+m_pos+1) == *(ush *)(in+ip+1)) {
			/*	We have found a match	*/
			uit look = len - ip - 2;
			if (look > MML) look = MML;
			m_len = 2;
			do {
				m_len++;
			} while(m_len != look && in[ip+m_len] == in[m_pos+m_len]);

			if (op + 2 + l_len + 3 >= maxlen) return - 1;
			if (l_len != 0) {
				out[op++] = (l_len - 1) << 3;
				do {
					out[op++] = in[ip-l_len--];
				} while (l_len != 0);
			}
			m_len -= 2;
			if (m_len <= 6) {
				out[op++] = m_len | ((m_off >> 5) & 0xf8);
			}
			else {
				out[op++] = 0x07 | ((m_off >> 5) & 0xf8);
				out[op++] = m_len - 7;
			}
			out[op++] = m_off & 0xff;
			ip++;
			hval = ((hval << 5) ^ in[ip+2]) & (HSIZ - 1);
			htab[hval] = ip;
			ip++;
			do {
				hval = ((hval << 5) ^ in[ip+2]) & (HSIZ - 1);
				htab[hval] = ip;
				ip++;
				m_len--;
			} while (0 != m_len);
		}
		else {
			/*	No match found	*/
			ip++;
			l_len++;
			if (MLL == l_len) {
				if (op + 2 + MLL >= maxlen) return -1;
				out[op++] = 0xf8;
				do {
					out[op++] = in[ip-l_len--];
				} while (l_len != 0);
			}
		}
	} while (ip < len);
	if (l_len != 0) {
		if (op + l_len + 3 >= maxlen) return -1;
		out[op++] = (l_len - 1) << 3;
		do {
			out[op++] = in[ip-l_len--];
		} while (l_len != 0);
	}
	*out_len = op;
	return 0;
}


#if !defined(LZV_ASM)

static lzo_uint
rLZV(const uch *in, uch *op, uit ilen, uit len)
{
	uit tbuf, c_len;
	const uch *ip, *in_end;
	uch *m_pos, *out_end, *o;

	ip = in;
	out_end = op + len;
	in_end = in + ilen;
	o = op;
	do {
		tbuf = *ip++;
		c_len = tbuf & 0x07;
		if (0 == c_len) {
			c_len = (tbuf >> 3) + 1;
			do *op++ = *ip++;
			while (--c_len);
		} else {
			if (0x07 == c_len) c_len = *ip++ + 7;
			m_pos = op - 1 - (((uit) (tbuf & 0xf8) << 5) | *ip++);
			*op++ = *m_pos++;
			*op++ = *m_pos++;
			do *op++ = *m_pos++;
			while (--c_len);
		}
	} while (op < out_end && ip < in_end);
	if (ip != in_end)
		return -1;
	/* return (ip-in); */
	return op - o;
}

#else
#error

/*
**	aCembler version
**
**	IMPORTANT: Use gcc -O2 -fomit-frame-pointer!!!
*/

static lzo_uint
rLZV(const uch *in, uch *out, uit ilen, uit len)
{
	register uit tbuf;
	register uch *ip;
	register uch *op asm ("%edi");
	register uch *m_pos asm ("%esi");
	uch *out_end, *in_end;
	register uit c_len asm ("%ecx");

	ip = in;
	op = out;
	out_end = out + len;
	in_end = in + ilen;
	asm volatile ("cld");
	do {
		tbuf = *ip++;
		c_len = tbuf & 0x07;
		if (0 == c_len) {
			c_len = (tbuf >> 3) + 1;
			m_pos = ip;
			ip += c_len;
			asm volatile ("rep; movsb"
				: /* no output */
				: "g" (m_pos), "g" (op), "g" (c_len));
		}
		else {
			if (0x07 == c_len) c_len = *ip++ + 9;
			else c_len += 2;
			m_pos = op - 1 - (((uit) (tbuf & 0xf8) << 5) | *ip++);
			asm volatile ("rep; movsb"
				: /* no output */
				: "g" (m_pos), "g" (op), "g" (c_len));
		}
	} while (op < out_end && ip < in_end);
	if(ip > in_end) return 0;
	return (ip-in);
}

#endif


/***********************************************************************
// LZV public entry points.
// return 0 for success, -1 for failure.
************************************************************************/

int
lzv_compress   ( const lzo_byte *in , lzo_uint  in_len,
                       lzo_byte *out, lzo_uint *out_len,
                       void *wrkmem )
{
	int r;

	/* sanity check */
	if (!lzo_assert(LZV_MEM_COMPRESS >= HSIZ*sizeof(htab_t)))
		return -1;

	r = wLZV(in,out,in_len,out_len,wrkmem);
	return r;
}


int 
lzv_decompress   ( const lzo_byte *in , lzo_uint  in_len,
                         lzo_byte *out, lzo_uint *out_len,
                         void *wrkmem /* NOT USED */ )
{
	int r;

	*out_len = rLZV(in,out,in_len,*out_len);
	r = (*out_len == (lzo_uint) -1) ? -1 : 0;
	return r;
}


/*
vi:ts=4
*/

